/* 	Copyright John D. "smeezekitty" 2017
	
	FreeNBTV is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeNBTV is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeNBTV.  If not, see <http://www.gnu.org/licenses/>.
*/

/*This file mostly handles the GUI for changing the NBTV generation parameters */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "nbtv_format.h"
#include "input.h"
#include "nbtv_gen.h"
#include "gui.h"


GtkWidget *format_grid;
GtkWidget *format_window;
GtkWidget *preset_selector;
GtkWidget *preset_save_button;
GtkWidget *preset_save_as_button;
GtkWidget *preset_delete_button;
GtkWidget *portrait,*landscape;
GtkWidget *orientation_box;
GtkWidget *ltr,*rtl;
GtkWidget *ttb,*btt;
GtkWidget *bright,*contrast;
GtkWidget *lines,*ppl;
GtkWidget *fps,*aspect;
GtkWidget *pixelck,*fblank;
GtkWidget *lblank,*fblank_val;
GtkWidget *lblank_val,*line_clock;
GtkWidget *line_edit, *channel_setup;
GtkWidget *first_usable,*last_usable;
GtkWidget *fpf,*fpf_conf;

extern GtkWidget *cs_window;

char f_handler_disable; //Disables the callback handlers to avoid recursion
char f_window_disable; //Disable the format window

//Fills in the preset selection drop down box
void populate_preset_combobox(GtkWidget *cb){
	int i;
	char name[64];
	f_handler_disable = 1;
	gtk_combo_box_text_remove_all(GTK_COMBO_BOX_TEXT(cb));
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), NULL, "Custom");
	for(i = 0;i < NUM_BUILT_IN_FORMATS;i++){
		sprintf(name, "%d %s", i+1, built_in_formats[i].preset_name);
		gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), NULL, name);
	}
	for(i = 0;i < last_user_fmt;i++){
		sprintf(name, "%d %s", i+1+NUM_BUILT_IN_FORMATS, user_formats[i].preset_name);
		gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(cb), NULL, name);
	}
	f_handler_disable = 0;
}
//Creates a dialog to accept the name for making a new preset format
void save_format_dialog(){
	GtkWidget *dialog;
	GtkWidget *entry_box;
	char *text;
	dialog = gtk_dialog_new_with_buttons("Save Format: Name", GTK_WINDOW(format_window), GTK_DIALOG_MODAL, "_OK", GTK_RESPONSE_ACCEPT, "_Cancel", GTK_RESPONSE_REJECT, NULL);
	entry_box = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(dialog))), entry_box, TRUE, TRUE, 0);
	gtk_widget_show(entry_box);
	if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT){
		text=(char*)gtk_entry_get_text(GTK_ENTRY(entry_box));
		debug("Accept: %s\n", text);
		if(!*text){
			error("Empty string\n");
			return;
		}
		if(strlen(text) > 64){
			error("Name too long!");
			return;
		}
		add_format(text);
	}
	gtk_widget_destroy(dialog);		
}
GtkWidget *create_preset_combobox(){
	GtkWidget *cb = gtk_combo_box_text_new();
	char name[40];
	int i;
	populate_preset_combobox(cb);
	set_NBTV_format(1);
	return cb;
}
void preset_save(GtkWidget *button, gpointer bla){
	debug("Clicked save\n");
	save_format_dialog();
	populate_preset_combobox(preset_selector);
	save_formats();
	gui_update_format(nbtv_fmt);
}
void preset_delete(GtkWidget *button, gpointer bla){
	debug("Clicked delete\n");
	delete_format();
	populate_preset_combobox(preset_selector);
	gui_update_format(nbtv_fmt);
	save_formats();
}
void start_line_editor(GtkWidget *button, gpointer bla){
	debug("Line editor\n");
}
//Open the channel setup window
void start_channel_setup(GtkWidget *button, gpointer bla){
	debug("Channel setup\n");
	if(!cs_window){
		create_channel_setup_window();
	} else {
		gtk_window_present(GTK_WINDOW(cs_window));
	}
	gui_update_channel_setup(nbtv_fmt);
}
//Preset selected from dropdown
void preset_select(GtkWidget *cb, gpointer bla){
	if(f_handler_disable) return;
	int n =  gtk_combo_box_get_active(GTK_COMBO_BOX(cb));
	set_NBTV_format(n);
	format_changed=1;
	gui_update_format(nbtv_fmt);
}
void orientation_select(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_orientation(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)));
}
void scan_direction_ltr(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_left_right_scan(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)));
}
void scan_direction_ttb(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_top_down_scan(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)));
}
void brightness_change(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_brightness(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void contrast_change(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_contrast(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void lines_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_lines(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void ppl_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_ppl(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void fps_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_framerate(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void pixel_clock_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_pixel_clock(atoi(gtk_entry_get_text(GTK_ENTRY(w))));
}
void usable_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_usable_lines(gtk_spin_button_get_value(GTK_SPIN_BUTTON(first_usable)), gtk_spin_button_get_value(GTK_SPIN_BUTTON(last_usable))); 
}
void fblank_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_frame_blank_time(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void lblank_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_line_blank_time(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void fblank_val_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_frame_blank_value(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void lblank_val_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_line_blank_value(gtk_spin_button_get_value(GTK_SPIN_BUTTON(w)));
}
void screen_aspect_changed(GtkWidget *w, gpointer bla){
	if(f_handler_disable) return;
	nbtv_format_set_screen_aspect(atof(gtk_entry_get_text(GTK_ENTRY(w))));
}
//Just update the drop down box
void gui_update_format_selector(){
	f_handler_disable = 1;
	gtk_combo_box_set_active(GTK_COMBO_BOX(preset_selector), format_index);
	f_handler_disable = 0;
}
//Update all fields to match the internal structure
void gui_update_format(struct NBTV_format *f){
	char buf[8];
	if(!format_window) return;
	f_handler_disable = 1;
	if(f_window_disable)
		gtk_widget_set_sensitive(format_grid, 0);
	else gtk_widget_set_sensitive(format_grid, 1);
	
	gtk_combo_box_set_active(GTK_COMBO_BOX(preset_selector), format_index);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(portrait), f->orientation);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ltr), f->left_right_scan);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ttb), f->top_down_scan);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(bright), f->brightness);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(contrast), f->contrast);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(lines), f->lines);
	gtk_entry_set_text(GTK_ENTRY(ppl), itoa(f->pixels_per_line, buf, 10));
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(fps), f->frame_rate);
	gtk_entry_set_text(GTK_ENTRY(aspect), gcvt(f->aspect_ratio, 4, buf));
	gtk_entry_set_text(GTK_ENTRY(pixelck), itoa(f->pixel_clock, buf, 10));
	gtk_entry_set_text(GTK_ENTRY(fblank), gcvt(f->frame_blanking_time, 4, buf));
	gtk_entry_set_text(GTK_ENTRY(lblank), gcvt(f->line_blanking_time, 4, buf));
	gtk_entry_set_text(GTK_ENTRY(fblank_val), itoa(f->frame_blanking_value, buf, 10));
	gtk_entry_set_text(GTK_ENTRY(lblank_val), itoa(f->line_blanking_value, buf, 10));
	sprintf(buf, "%0.1f", f->line_clock);
	gtk_label_set_text(GTK_LABEL(line_clock), buf);
	gtk_spin_button_set_range(GTK_SPIN_BUTTON(first_usable), 0, f->lines - 1);
	gtk_spin_button_set_range(GTK_SPIN_BUTTON(last_usable), 0, f->lines - 1);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(first_usable), f->first_usable_line);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(last_usable), f->last_usable_line);
	f_handler_disable = 0;
}
//Create the NBTV format window
GtkWidget *create_format_window(){
	format_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect (format_window, "destroy",G_CALLBACK (destroy), NULL);
	gtk_window_set_title(GTK_WINDOW(format_window), "Format");
	gtk_window_set_resizable(GTK_WINDOW(format_window), FALSE);
	format_grid = gtk_grid_new();
	gtk_grid_set_row_homogeneous(GTK_GRID(format_grid), TRUE);
	gtk_grid_set_row_spacing(GTK_GRID(format_grid), 6);
	gtk_container_add(GTK_CONTAINER(format_window), format_grid);
	preset_selector = create_preset_combobox();
	gtk_grid_attach(GTK_GRID(format_grid), preset_selector, 1, 0, 3, 1);
	preset_save_button = gtk_button_new_with_label("Save");
	g_signal_connect(preset_save_button, "clicked", G_CALLBACK(preset_save), NULL);
	gtk_grid_attach(GTK_GRID(format_grid), preset_save_button, 1, 1, 1, 1);
	//preset_save_as_button = gtk_button_new_with_label("Save As");
	//g_signal_connect(preset_save_as_button, "clicked", G_CALLBACK(preset_save_as), NULL);
	gtk_grid_attach(GTK_GRID(format_grid), preset_save_as_button, 2, 1, 1, 1);
	preset_delete_button = gtk_button_new_with_label("Delete");
	g_signal_connect(preset_delete_button, "clicked", G_CALLBACK(preset_delete), NULL);
	gtk_grid_attach(GTK_GRID(format_grid), preset_delete_button, 3, 1, 1, 1);
	gtk_grid_attach(GTK_GRID(format_grid), gtk_label_new("Orientation:"), 1, 2, 1, 1);
	portrait = gtk_radio_button_new_with_label(NULL, "Portrait");
	gtk_grid_attach(GTK_GRID(format_grid), portrait, 2, 2, 1, 1);
	landscape = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(portrait), "Landscape");
	gtk_grid_attach(GTK_GRID(format_grid), landscape, 3, 2, 1, 1);
	orientation_box = gtk_frame_new("Scan direction: ");
	gtk_grid_attach(GTK_GRID(format_grid), orientation_box, 1, 3, 3, 3);
	rtl = gtk_radio_button_new_with_label(NULL, "Left<-Right");
	ltr = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(rtl), "Left->Right");
	gtk_grid_attach(GTK_GRID(format_grid), ltr, 1, 4, 1, 1);
	gtk_grid_attach(GTK_GRID(format_grid), rtl, 3, 4, 1, 1);
	btt = gtk_radio_button_new_with_label(NULL, "Bottom->Top");
	ttb = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(btt), "Top->Bottom");
	gtk_grid_attach(GTK_GRID(format_grid), ttb, 1, 5, 1, 1);
	gtk_grid_attach(GTK_GRID(format_grid), btt, 3, 5, 1, 1);
	gtk_grid_attach(GTK_GRID(format_grid), gtk_label_new("Preset: "), 0, 0, 1, 1);
	put_label(format_grid, "Brightness: ", 0, 6);
	put_label(format_grid, "Contrast: ", 3, 6);
	put_label(format_grid, "Lines: ", 0, 7);
	put_label(format_grid, "Pixels\nPer Line: ", 0, 8);
	put_label(format_grid, "Frame\nRate (FPS): ", 0, 9);
	put_label(format_grid, "Physical\nAspect (w/h): ", 0, 10);
	put_label(format_grid, "Pixel\nClock (Hz): ", 0, 11);
	put_label(format_grid, "Line Clock (Hz): ", 3, 11);
	put_label(format_grid, "Frame\nBlanking (\xC2\xB5S)" /* uS */, 3, 7);
	put_label(format_grid, "Line\nBlanking (\xC2\xB5S)" /* uS */, 3, 8);
	put_label(format_grid, "F Blanking\nValue (% F/S)", 3, 9);
	put_label(format_grid, "L Blanking\nValue (% F/S)", 3, 10);
	put_label(format_grid, "Fields/Frame:", 0, 13);
	bright = gtk_spin_button_new_with_range(-100, 100, 1.0);
	gtk_grid_attach(GTK_GRID(format_grid), bright, 1, 6, 1, 1);
	contrast = gtk_spin_button_new_with_range(-100, 100, 1.0);
	gtk_grid_attach(GTK_GRID(format_grid), contrast, 4, 6, 1, 1);
	lines = gtk_spin_button_new_with_range(4, 240, 1.0);
	gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(lines), TRUE);
	gtk_grid_attach(GTK_GRID(format_grid), lines, 1, 7, 1, 1);
	ppl = gtk_spin_button_new_with_range(4, 360, 1.0);
	gtk_grid_attach(GTK_GRID(format_grid), ppl, 1, 8, 1, 1);
	fps = gtk_spin_button_new_with_range(0.5, 90, 0.05);;
	gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(fps), TRUE);
	gtk_grid_attach(GTK_GRID(format_grid), fps, 1, 9, 1, 1);
	aspect = create_number_box(4);
	gtk_grid_attach(GTK_GRID(format_grid), aspect, 1, 10, 1, 1);
	pixelck = create_number_box(6);
	gtk_grid_attach(GTK_GRID(format_grid), pixelck, 1, 11, 1, 1);
	fblank = gtk_spin_button_new_with_range(0, 99999, 0.5);
	gtk_grid_attach(GTK_GRID(format_grid), fblank, 4, 7, 1, 1);
	lblank = gtk_spin_button_new_with_range(0, 1000, 0.5);
	gtk_grid_attach(GTK_GRID(format_grid), lblank, 4, 8, 1, 1);
	fblank_val = gtk_spin_button_new_with_range(0, 100, 1);
	gtk_grid_attach(GTK_GRID(format_grid), fblank_val, 4, 9, 1, 1);
	lblank_val = gtk_spin_button_new_with_range(0, 100, 1);
	gtk_grid_attach(GTK_GRID(format_grid), lblank_val, 4, 10, 1, 1);
	line_clock = gtk_label_new("400");
	gtk_grid_attach(GTK_GRID(format_grid), line_clock, 4, 11, 1, 1);
	channel_setup = gtk_button_new_with_label("Channel Setup");
	g_signal_connect(channel_setup, "clicked", G_CALLBACK(start_channel_setup), NULL);
	gtk_grid_attach(GTK_GRID(format_grid), channel_setup, 0, 14, 2, 1);
	line_edit = gtk_button_new_with_label("Line Editor");
	gtk_widget_set_sensitive(line_edit, FALSE);
	g_signal_connect(line_edit, "clicked", G_CALLBACK(start_line_editor), NULL);
	gtk_grid_attach(GTK_GRID(format_grid), line_edit, 3, 14, 2, 1);
	first_usable = gtk_spin_button_new_with_range(0, 31, 1.0);
	last_usable = gtk_spin_button_new_with_range(0, 31, 1.0);
	gtk_grid_attach(GTK_GRID(format_grid), gtk_label_new("Usable lines\n(0 - (n-1)): "), 0, 12, 1, 1);
	gtk_grid_attach(GTK_GRID(format_grid), first_usable, 1, 12, 1, 1);
	gtk_grid_attach(GTK_GRID(format_grid), last_usable, 4, 12, 1, 1);
	gtk_grid_attach(GTK_GRID(format_grid), gtk_label_new("\xE2\x80\x94\xE2\x80\x94TO\xE2\x80\x94\xE2\x80\x94") /* -- TO -- */, 2, 12, 2, 1);
	fpf = gtk_spin_button_new_with_range(1, 4, 1);
	gtk_widget_set_sensitive(fpf, FALSE);
	gtk_grid_attach(GTK_GRID(format_grid), fpf, 1, 13, 1, 1);
	fpf_conf = gtk_button_new_with_label("Configure");
	gtk_widget_set_sensitive(fpf_conf, FALSE);
	gtk_grid_attach(GTK_GRID(format_grid), fpf_conf, 4, 13, 1, 1);
	//put_label(format_grid, "\xE2\x80\x94TO\xE2\x80\x94", 2, 12);
	gtk_widget_show_all(format_window);
}
void gui_format_connect_callbacks(){
	g_signal_connect(preset_selector, "changed", G_CALLBACK(preset_select), NULL);
	g_signal_connect(portrait, "toggled", G_CALLBACK(orientation_select), NULL);
	g_signal_connect(ltr, "toggled", G_CALLBACK(scan_direction_ltr), NULL);
	g_signal_connect(ttb, "toggled", G_CALLBACK(scan_direction_ttb), NULL);
	g_signal_connect(bright, "value-changed", G_CALLBACK(brightness_change), NULL);
	g_signal_connect(contrast, "value-changed", G_CALLBACK(contrast_change), NULL);
	g_signal_connect(lines, "value-changed", G_CALLBACK(lines_changed), NULL);
	g_signal_connect(fps, "value-changed", G_CALLBACK(fps_changed), NULL);
	g_signal_connect(ppl, "value-changed", G_CALLBACK(ppl_changed), NULL);
	g_signal_connect(aspect, "changed", G_CALLBACK(screen_aspect_changed), NULL);
	g_signal_connect(pixelck, "changed", G_CALLBACK(pixel_clock_changed), NULL);
	g_signal_connect(first_usable, "value-changed", G_CALLBACK(usable_changed), NULL);
	g_signal_connect(last_usable, "value-changed", G_CALLBACK(usable_changed), NULL);
	g_signal_connect(fblank, "value-changed", G_CALLBACK(fblank_changed), NULL);
	g_signal_connect(lblank, "value-changed", G_CALLBACK(lblank_changed), NULL);
	g_signal_connect(fblank_val, "value-changed", G_CALLBACK(fblank_val_changed), NULL);
	g_signal_connect(lblank_val, "value-changed", G_CALLBACK(lblank_val_changed), NULL);
}
void gui_format_disable_window(){
	if(f_window_disable) return;
	f_window_disable = 1;
	gui_update_format(nbtv_fmt);
}
void gui_format_enable_window(){
	if(!f_window_disable) return;
	f_window_disable = 0;
	gui_update_format(nbtv_fmt);
}